
#include "scan_job_manager.h"
#include "scan_job_states.h"
#include "scanner_module.h"

#include "scanner_ui/pending_commands_queue.h"
#include "scanner_ui/cmd_handlers.h"
#include "scanner_ui/scanner_output_queue.h"
#include "scanner_ui/scanner_ui.h"

#include "packet_submitter.h"
#include "scan_jobs.h"
#include "queue.h"

#include "sniffer/packet_receiver.h"


#include <linux/kthread.h>

struct task_struct *scan_job_man_thread;

/* mutex used to make sure the module doesn't
 * unload itsself before the kernel-thread has
 * been deinitialized and has finished.
 */

static __DECLARE_SEMAPHORE_GENERIC(fini_mutex, 0);

/** \addtogroup ScanJobManager
    @{ 
*/

/** hash-table which maps command-names to handler-functions */
struct cmd_handlers_hash_bucket cmd_handlers_hash[CMD_HASH_SIZE];

/** @name Private functions
    @{    
*/

/**
   polls the sniffed-packet-queue.
      
*/


static void process_pending_packets(void)
{
	
	/* poll the global 'sniffed-packet-queue' provided by the
	   packet-receiver
	*/
		
	struct sniffed_packet_queue_node *packet_queue =
		packet_receiver.sniffed_packet_queue;
       	
	struct sniffed_packet_descr_t *packet_descr;
		
	
	/* for each element in packet-queue: */
	while((packet_descr = sniffed_packet_queue_head(packet_queue))){
		
		/* determine sender and forward the packet
		   to the scan-job which handles this host. */
		struct scan_job_t *sj;
		
		sj = scan_jobs_hash_get(packet_descr->subject);				
		
		if(!sj){			
			kfree(packet_descr);
			continue;
		}
				
		/* Insert packet into the queue of the scan-job */
		sniffed_packet_queue_add(&sj->packets, packet_descr);		
		
	}
		
	
}

/**
   
   For each scan-job in the active_scan_jobs_list,
   execute its scan_job_manager-function.
   
   The scan_job_manager-function's return value will
   indicate whether the scan-job has finished or needs
   to be called again. If it has finished, this function
   removes the scan_job from the scan_jobs_hash and it
   will not be reinserted into the list of
   active scan-jobs.
     
   Return the next time scan-jobs must be processed due
   to timeouts.

*/

static s64 process_active_scan_jobs(void)
{	
	s64 timeout_time;
	s64 smallest_timeout_time = 0;	
	
	struct list_head *p, *n;
	list_for_each_safe(p, n, &active_scan_jobs_list.list){
		struct scan_jobs_node *entry
			= list_entry(p, struct scan_jobs_node, list);
		
		struct scan_job_t *cur_scan_job = 
			entry->scan_job;
				

		/* execute scan-job-manager for the current state */
		
		timeout_time = scan_job_states[cur_scan_job->state]->scan_job_manager(cur_scan_job);
		
		if(timeout_time != FINISHED){
			if(smallest_timeout_time == 0 || timeout_time < smallest_timeout_time)
				smallest_timeout_time = timeout_time;		
			continue;
		}
				

		/* remove scan-job from active-scan-jobs-list */
		entry->scan_job->active = FALSE;
		list_del(&entry->list);
		kfree(entry);
		
		scanner_output_queue_add("R SCAN_JOB_REMOVED ");
		scanner_output_queue_add(cur_scan_job->addr_str);		
		
		/* remove from hash as well. */
		
		scan_jobs_hash_remove(cur_scan_job->addr);		
		scanner_output_queue_add("\n");
		
		//return;	
	}			

	return smallest_timeout_time;
	
}

/**
   Handle pending commands

   Polls the pending_commands_queue for commands which were
   received by the scanner-userland-interface and executes
   the associated command-handler-function.  

   Returns TRUE if the scanner should shutdown,
   FALSE otherwise

*/

static boolean process_pending_commands(void)
{
	struct command_t *cmd;	
	
	/* if there are no active scan-jobs, it's allowed
	 * to sleep in the pending_commands_queue_head-call.
	 */
	boolean may_sleep = list_empty(&active_scan_jobs_list.list);

	while((cmd = pending_commands_queue_head(may_sleep))){
		
		
		if(cmd->name && (strcmp(cmd->name, "exit") == 0)){
			delete_command(cmd);
			return TRUE;
		}
			
		printk("cmd->name: %s\n", cmd->name);		
				
		
		cmd_handlers_execute(cmd_handlers_hash, cmd);
		
		delete_command(cmd);
	
		may_sleep = list_empty(&active_scan_jobs_list.list);
			
	}

	return FALSE;
}


/**
  
  \brief  kernel-thread's main function.
  
  This function will call process_pending_packets,
  process_submitter_reports, process_active_scan_jobs
  and process_pending_commands in a loop until it is
  asked to terminate. 
  
*/

static int scan_job_manager(void *unused)
{		

	daemonize("scan-job-manager");
	allow_signal(SIGINT);
	printk("Scan-Job-Manager started\n");
	
	while(!kthread_should_stop()){
		boolean stop;
		s64 next_timeout;
		struct timespec cur_time;

		process_pending_packets();
		next_timeout = process_active_scan_jobs();		
		stop = process_pending_commands();		
		
		if(stop) break;
				
						
		/* check if it makes sense to sleep for a millisecond */
		
		
		getnstimeofday(&cur_time);
		if( next_timeout - timespec_to_ns(&cur_time) <= 1000000)		
			continue;
		
	
		// sleep but allow interruption.
		msleep_interruptible(1);				
	}

		
	scan_job_manager_fini();
	scan_job_man_thread = NULL;
	printk("Shutting down scan-job-manager\n");
	
	up(&fini_mutex);
	return 0;
}




/** @} */

int scan_job_manager_init(void)
{
	int ret = scan_job_states_init();
	if(ret != SUCCESS)
		return ret;
	
	cmd_handlers_init(cmd_handlers_hash);
	
	/* now register all cmd-handlers */

	cmd_handlers_register(cmd_handlers_hash, "create_scanjob", &handle_create_scanjob );
	cmd_handlers_register(cmd_handlers_hash, "execute_scanjob", &handle_execute_scanjob );
	cmd_handlers_register(cmd_handlers_hash, "pause_scanjob", &handle_pause_scanjob) ;
	cmd_handlers_register(cmd_handlers_hash, "remove_scanjob", &handle_remove_scanjob);
	
	scan_jobs_hash_init();

	scan_job_man_thread = kthread_run(&scan_job_manager, NULL, "scan-job-manager");
	
	if(scan_job_man_thread == ERR_PTR(-ENOMEM))
		return FAILURE;


	return SUCCESS;
}


static void clear_pending_packets(void)
{
	
	/* poll the global 'sniffed-packet-queue' provided by the
	   packet-receiver
	*/
		
	struct sniffed_packet_queue_node *packet_queue =
		packet_receiver.sniffed_packet_queue;
       	
	struct sniffed_packet_descr_t *packet_descr;
		
	
	/* for each element in packet-queue: */
	while((packet_descr = sniffed_packet_queue_head(packet_queue))){		
		kfree(packet_descr);				
	}
		
	
}


void scan_job_manager_fini(void)
{
  	
	/* first make sure no new
	 * commands or packets reach
	 * the scan-job-manager
	 */
	scanner_ui_fini();	
  
	packet_receiver_fini();	
	
	/* Now it's safe to deinit
	   the scan-job-manager
	*/	
	
	clear_pending_packets();
	active_scan_jobs_list_fini();
	scan_jobs_hash_fini();
	cmd_handlers_fini(cmd_handlers_hash);
	scan_job_states_fini();	

}


void scan_job_manager_wait_for_termination(void)
{
	down(&fini_mutex);
}

/** @} */  /* End ScanJobManager-Group */

